Basic algorithms of compensated arithmetic.
Most of the basic numerical algorithms are directly lifted from:
- M. Joldes, et al., ACM Trans. Math. Softw. 44, 1-27 (2018)
- Karp, High Precision Division and Square Root (1993)
Basic algorithms of compensated arithmetic.
Most of the basic numerical algorithms are directly lifted from: